/**
 ** sahandMainCodels.cc
 **
 ** Codels called by execution task sahandMain
 **
 ** Author: 
 ** Date: 
 **
 **/

#include <math.h>

#include <portLib.h>
#include <h2mathLib.h>

#include <SAHandCtrlApi.h>

#include "server/sahandHeader.h"
#include "sahandConst.h"
#include "sahandPrivate.h"

static CSAHandCtrlApi sahapi;		// Schunk API object
static POSTER_ID trackPosterId = NULL;

/*----------------------------------------------------------------------*/
/* Convert a Genom port number/index to a Schunk API port */
static int
sahandPort(int p)
{
	switch (p) {
	case SAH_PORT_1: return PORT_1;
	case SAH_PORT_2: return PORT_2;
	default: return -1;
	}
}
/*----------------------------------------------------------------------*/
static STATUS
sahandOneInit(int hand, int *report)
{
	float pos[3], vel[3], stiffness[3];
	SAHAND_FINGER_REF *f;
	int i, j, angle;
	int port = sahandPort(hand);
	short res;

	/* Default speeds */
	for (i = 0; i < SAH_MAX_FINGERS; i++) {
		f = &(SDI_F->sah_ref[hand].sr_fingers[i]);
		for (j = 0; j < 3; j++)
			f->sfr_vel[j] = DEG_TO_RAD(100.0);
	}
	/* reset controller mode */
	res = sahapi.SetController(port, SAH_CONTROLLER_RESET);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	/* Controller defaults to IMPEDANCE */
	res = sahapi.SetController(port, SAH_CONTROLLER_IMPEDANCE);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	SDI_F->sah_params[hand].sap_controller = SAH_CONTROLLER_IMPEDANCE;

	/* Release brakes */
	res = sahapi.SetEmergencyStop(port, RUN);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	SDI_F->sah_state[hand].ss_brake = SAH_RUN;
	res = sahapi.SetThumbBrake(port, RUN);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	SDI_F->sah_state[hand].ss_thumb.sts_brake = SAH_RUN;

	/* Set all joints references to current pos, and 0 stiffness */
	for (i = 0; i < SAH_MAX_FINGERS; i++) {
		f = &(SDI_F->sah_ref[hand].sr_fingers[i]);
		for (j = 0; j < 3; j++) {
			vel[j] = RAD_TO_DEG(f->sfr_vel[j]);
			stiffness[j] = 0.0;
		}
		sahapi.SetFingerEnable(port, i, TRUE);
		sahapi.GetJointAngle(port, i, pos);
		/* Set reference to current position */
		for (j = 0; j < 3; j++)
			f->sfr_angle[j] = DEG_TO_RAD(pos[j]);
		sahapi.MoveFinger(port, i, pos, vel, 0.0);
		sahapi.SetStiffnessFactor(port, i, stiffness);
	}
	
	/* Update thumb reference  */
	sahapi.GetThumbAngle(port, &angle);
	SDI_F->sah_ref[hand].sr_thumb_orientation = DEG_TO_RAD((double)angle);
	return OK;
}
/*------------------------------------------------------------------------
 *
 * sahandInit  --  Initialization codel (fIDS, ...)
 *
 * Description: connects to the driver, sets all parameters to a known state.
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandInit(int *report)
{
  int i, config, port;
  short res;
  
  res = sahapi.SAHandInit();
  if (res <= 0) {
    *report = S_sahand_DRIVER_ERROR;
    return ERROR;
  }
  for (i = 0; i < 2; i++) {
    port = sahandPort(i);
    /* Is it a left or right hand ? */
    res = sahapi.GetHandConfig(port, &config);
    if (res <= 0) {
      /* error */
      *report = S_sahand_DRIVER_ERROR;
      return ERROR;
    } else {
      switch(config) {
      case 0:
	SDI_F->sah_state[i].ss_config = 
	  SAH_CONFIG_NOT_PRESENT;
	break;
      case HAND_CONFIG_LEFT:
	SDI_F->sah_state[i].ss_config = 
	  SAH_CONFIG_LEFT;
	break;
      case HAND_CONFIG_RIGHT:
	SDI_F->sah_state[i].ss_config = 
	  SAH_CONFIG_RIGHT;
	break;
      default:
	*report = S_sahand_DRIVER_ERROR;
	return ERROR;
      } /* case */
      /* init this hand */
      if (sahandOneInit(i, report) == ERROR)
	return ERROR;
    }
  } /* for */
  sahandInitCheck();
  SDI_F->posterHand.cpt = 0;
  return OK;
}

/*------------------------------------------------------------------------
 *
 * sahandEnd  --  Termination codel
 *
 * Description: puts brakes on, disconnect from the driver.
 *
 * Returns:    OK or ERROR
 */

STATUS
sahandEnd(void)
{
	int i, port;
	
	for (i = 0; i < 2; i++) {
		if (!SAHAND_PORT_ACTIVE(i)) 
			continue;
		port = sahandPort(i);
		sahapi.SetEmergencyStop(port, TRUE);
		sahapi.SetThumbBrake(port, TRUE);
	}
	return OK;
}

/*----------------------------------------------------------------------*/
/**
 ** Helper functions for the permanent activity
 **
 ** Wrapping the API provided by Schunk
 **/

static short
sahandSetController(int hand)
{
	int controller = SDI_F->sah_params[hand].sap_controller,
	    port = sahandPort(hand);

	return sahapi.SetController(port, controller);
}

static short
sahandSetThumbAngle(int hand)
{
	int angle = (int)RAD_TO_DEG(SDI_F->sah_ref[hand].sr_thumb_orientation);
	int port = sahandPort(hand);

	return sahapi.SetThumbAngle(port, angle);
}

static short
sahandSetThumbBrake(int hand)
{
	return 0;
}

static short
sahandMoveHand(int hand)
{
	SAH_DESIRED desired;
	SAHAND_FINGER_REF *finger;
	SAHAND_STIFFNESS_FACTOR *ssf;
	int i, j, port = sahandPort(hand);

	finger = &SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_THUMB];
	ssf = &SDI_F->sah_params[hand].sap_stiffness[SAH_FINGER_THUMB];
	/* Convert parameters for the Schunk API */
	for (i = 0; i < 3; i++) {
		desired.Thumb.Angle[i] =
		    RAD_TO_DEG(finger->sfr_angle[i]);
		desired.Thumb.Velocity[i] =
		    RAD_TO_DEG(finger->sfr_vel[i]);
		desired.Thumb.Stiffness[i] = ssf->ssf_value[i];
	}
	finger = &SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_FIRST];
	ssf = &SDI_F->sah_params[hand].sap_stiffness[SAH_FINGER_FIRST];
	for (i = 0; i < 3; i++) {
		desired.FirstFinger.Angle[i] =
		    RAD_TO_DEG(finger->sfr_angle[i]);
		desired.FirstFinger.Velocity[i] =
		    RAD_TO_DEG(finger->sfr_vel[i]);
		desired.FirstFinger.Stiffness[i] = ssf->ssf_value[i];
	}
	finger = &SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_MIDDLE];
	ssf = &SDI_F->sah_params[hand].sap_stiffness[SAH_FINGER_MIDDLE];
	for (i = 0; i < 3; i++) {
		desired.MiddleFinger.Angle[i] =
		    RAD_TO_DEG(finger->sfr_angle[i]);
		desired.MiddleFinger.Velocity[i] =
		    RAD_TO_DEG(finger->sfr_vel[i]);
		desired.MiddleFinger.Stiffness[i] = ssf->ssf_value[i];
	}
	finger = &SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_RING];
	ssf = &SDI_F->sah_params[hand].sap_stiffness[SAH_FINGER_RING];
	for (i = 0; i < 3; i++) {
		desired.RingFinger.Angle[i] =
		    RAD_TO_DEG(finger->sfr_angle[i]);
		desired.RingFinger.Velocity[i] =
		    RAD_TO_DEG(finger->sfr_vel[i]);
		desired.RingFinger.Stiffness[i] = ssf->ssf_value[i];
	}
	return sahapi.MoveHand(port, &desired);
}

static short
sahandSetStiffness(int hand)
{
	double *ref;
	float stiffness[3];
	int i, j;
	int port = sahandPort(hand);
	short res;
	
	for (i = 0; i < SAH_MAX_FINGERS; i++) {
		ref = &SDI_F->sah_params[hand].sap_stiffness[i].ssf_value[0];
		for (j = 0; j < 3; j++)
			stiffness[j] = (float)ref[j];
		res = sahapi.SetStiffnessFactor(port, i, stiffness);
		if (res <= 0) 
			break;
	}
	return res;
}

/*----------------------------------------------------------------------*/

/**
 ** Send pending action commands to the driver
 **/
static STATUS
sahandAction(int hand, int *report)
{
	int command = SDI_F->sah_command[hand];
	short res;

	/* Pending actions are set in the control task */
	if (command & SAH_COMMAND_SET_CONTROLLER)
		res = sahandSetController(hand);
	if (command & SAH_COMMAND_SET_THUMB_ANGLE)
		res = sahandSetThumbAngle(hand);
	if (command & SAH_COMMAND_SET_THUMB_BRAKE)
		res = sahandSetThumbBrake(hand);
	if (command & SAH_COMMAND_MOVE_HAND)
		res = sahandMoveHand(hand);
	if (command & SAH_COMMAND_SET_STIFFNESS_FACTOR)
		res = sahandSetStiffness(hand);
	/* Store last result */
	SDI_F->sah_state[hand].ss_error = res;
	/* Clear pending actions */
	SDI_F->sah_command[hand] = 0; 		
	return OK;
}

/*----------------------------------------------------------------------*/
/**
 ** Read sensor values
 **/
static STATUS
sahandSensors(int hand, int *report)
{
	float temp, pos[3];
	SAHAND_FINGER_STATE *f;
	int i, j, tpos, port = sahandPort(hand);
	short res;

	/* Fingers */
	for (i = 0; i < SAH_MAX_FINGERS; i++) {
		f = &(SDI_F->sah_state[hand].ss_finger[i]);
		/* Position */
		res = sahapi.GetJointAngle(port, i, pos);
		if (res <= 0) {
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		}
		for (j = 0; j < 3; j++)
			f->sfs_pos[j] = DEG_TO_RAD(pos[j]);
		/* speed */
		res = sahapi.GetJointSpeed(port, i, pos);
		if (res <= 0) {
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		}
		for (j = 0; j < 3; j++)
			f->sfs_vel[j] = DEG_TO_RAD(pos[j]);
		/* torque */
		res = sahapi.GetJointTorque(port, i, pos);
		if (res <= 0) {
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		}
		for (j = 0; j < 3; j++)
			f->sfs_torque[j] = pos[j];
		/* Finger temperature */
		res = sahapi.GetFingerTemp(port, i, &temp);
		if (res <= 0) {
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		}
		f->sfs_finger_temp = (double)temp;
		/* Joint limits */
		res = sahapi.GetFingerLimitsStatus(port, i, f->sfs_limits);
		if (res <= 0) {
			*report = S_sahand_DRIVER_ERROR;
			return ERROR;
		}
		
	}
	/* Thumb position */
	res = sahapi.GetThumbAngle(port, &tpos);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	SDI_F->sah_state[hand].ss_thumb.sts_angle = DEG_TO_RAD((double)tpos);

	/* Palm temperature */
	res = sahapi.GetPalmTemp(port, &temp);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return ERROR;
	}
	SDI_F->sah_state[hand].ss_palm_temp = temp;
	return OK;
}

/*------------------------------------------------------------------------
 *
 * sahandMainPerm  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 *              S_sahand_DRIVER_ERROR
 * 
 * Returns:    OK or ERROR
 */

STATUS
sahandMainPerm(int *report)
{
	int hand;

	for (hand = 0; hand < 2; hand++) {
		if (!SAHAND_PORT_ACTIVE(hand)) 
			continue;

		/* Sensors */
		if (sahandSensors(hand, report) == ERROR)
			return ERROR;
	
		/* Actions */
		if (SDI_F->sah_command[hand] != 0) 
			if (sahandAction(hand, report) == ERROR)
				return ERROR;
	}
	return OK;
}
/*------------------------------------------------------------------------
 * ClearTorqueOffset
 *
 * Description: clear existing offset on sensors for one finger
 *
 * Reports:      OK
 *              S_sahand_DRIVER_ERROR
 *              S_sahand_INACTIVE_PORT
 */

/* sahandClearTorqueOffset  -  codel EXEC of ClearTorqueOffset
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandClearTorqueOffset(SAHAND_THIS_FINGER *sah_finger_id, int *report)
{
	short res;

	res = sahapi.ClearTorqueSensorOffset(
	    sahandPort(sah_finger_id->stf_port), sah_finger_id->stf_id);
	SDI_F->sah_state[sah_finger_id->stf_port].ss_error = res;
	
	return ETHER;
}

/*------------------------------------------------------------------------
 * Brakes
 *
 * Description: Set brakes
 *
 * Reports:      OK
 *              S_sahand_DRIVER_ERROR
 *              S_sahand_INACTIVE_PORT
 */

/* sahandBrake  -  codel EXEC of Brakes
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandBrake(SAHAND_THIS_HAND_BRAKE *sah_brake, int *report)
{
	short res;

	res = sahapi.SetEmergencyStop(sah_brake->sthb_port, 
	    sah_brake->sthb_brake);
	if (res <= 0) {
		*report = S_sahand_DRIVER_ERROR;
		return FAIL;
	}
	return ETHER;
}

/*------------------------------------------------------------------------
 * TestGrasp
 *
 * Description: 
 *
 * Reports:      OK
 */

static int testCompteur;
static double testOrigin[SAH_MAX_FINGERS];
/* sahandTestGraspStart  -  codel START of TestGrasp
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandTestGraspStart(int *report)
{
  SAHAND_HAND_CLOSE *para;
  SAHAND_HAND_CLOSE_I *parai;
  int phalanx;
  int finger;
  int hand;

  hand = 0;
  if (!SAHAND_PORT_ACTIVE(hand)) {
    *report = S_sahand_INACTIVE_PORT;
    return ETHER;
  }
  testCompteur = 0;
  phalanx = SAHAND_PROXIMAL;

  finger =  SAH_FINGER_THUMB;
  testOrigin[finger] = SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[phalanx];
  finger =  SAH_FINGER_FIRST;
  testOrigin[finger] = SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[phalanx];
  finger =  SAH_FINGER_MIDDLE;
  testOrigin[finger] = SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[phalanx];
  finger =  SAH_FINGER_RING;
  testOrigin[finger] = SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[phalanx];
  return EXEC;
}

/* sahandTestGraspMain  -  codel EXEC of TestGrasp
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandTestGraspMain(int *report)
{
  int phalanx;
  int finger;
  int hand;
  double amplitude = (SAH_MAX_PROXIMAL_ANGLE - SAH_MIN_PROXIMAL_ANGLE) / 10;

  testCompteur++;
  hand = 0;
  if (!SAHAND_PORT_ACTIVE(hand)) {
    *report = S_sahand_INACTIVE_PORT;
    return ETHER;
  }
  phalanx = SAHAND_PROXIMAL;
  finger =  SAH_FINGER_THUMB;
  SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[phalanx] =
    testOrigin[finger] + sin(testCompteur*0.05) * amplitude;
  sahandFingerPositionLimit(&(SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[phalanx]));

  finger =  SAH_FINGER_FIRST;
  SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[phalanx] =
    testOrigin[finger] + sin(testCompteur*0.05) * amplitude;
  sahandFingerPositionLimit(&(SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[phalanx]));

  finger =  SAH_FINGER_MIDDLE;
  SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[phalanx] =
    testOrigin[finger] + sin(testCompteur*0.05 + 3.14) * amplitude;
  sahandFingerPositionLimit(&(SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[phalanx]));

  finger =  SAH_FINGER_RING;
  SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[phalanx] =
    testOrigin[finger] + sin(testCompteur*0.05) * amplitude;
  sahandFingerPositionLimit(&(SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[phalanx]));


  SDI_F->sah_command[hand] |= SAH_COMMAND_MOVE_HAND;
  return EXEC;
}

/*------------------------------------------------------------------------
 * GraspBottle
 *
 * Description: 
 *
 * Reports:      OK
 *              S_sahand_GRASP_ERROR
 */

/* sahandGraspBottleStart  -  codel START of GraspBottle
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandGraspBottleStart(int *report)
{
  SAHAND_HAND_CLOSE *para;
  SAHAND_HAND_CLOSE_I *parai;
  int phalanx;
  int finger;
  int hand;

  for (hand = 0; hand < 2; hand++) {
    if (!SAHAND_PORT_ACTIVE(hand)) 
      continue;
    para = &(SDI_F->sah_close[hand]);
    parai = &(SDI_F->sah_closei[hand]);
    // hypothesis : the additional thumb axis is not used
    for (finger = 0; finger < SAH_MAX_FINGERS; finger++) {
      // position and velocity
      parai->fingersI[finger].state[SAHAND_DISTAL] = SAHAND_FINGER_INACTIVE;
      SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_DISTAL] =
	SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[SAHAND_DISTAL];
      SDI_F->sah_ref[hand].sr_fingers[finger].sfr_vel[SAHAND_DISTAL] = 0;

      parai->fingersI[finger].state[SAHAND_PROXIMAL] = SAHAND_FINGER_CLOSING;
      SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_PROXIMAL] =
	para->fingers[finger].positionEnd[SAHAND_PROXIMAL];
      SDI_F->sah_ref[hand].sr_fingers[finger].sfr_vel[SAHAND_PROXIMAL] =
	para->fingers[finger].velocity[SAHAND_PROXIMAL];

      parai->fingersI[finger].state[SAHAND_BASE] = SAHAND_FINGER_INACTIVE;
      SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_BASE] =
	SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[SAHAND_BASE];
      SDI_F->sah_ref[hand].sr_fingers[finger].sfr_vel[SAHAND_BASE] = 0.;

      // stiffness and forces
      for (phalanx = 0; phalanx < 3; phalanx++) {
	/* memorize force to compensate gravity */
	parai->fingersI[finger].force0[phalanx] =
	  SDI_F->sah_state[hand].ss_finger[finger].sfs_torque[phalanx];
	/* We suppose verification have been done in Cntrl request ! */
	SDI_F->sah_params[hand].sap_stiffness[finger].ssf_value[phalanx] =
	  para->fingers[finger].stiffness[phalanx];
      }
    }
    parai->stateHand = SAHAND_CLOSING_PROXIMAL;
    SDI_F->sah_command[hand] |= SAH_COMMAND_MOVE_HAND;
    SDI_F->sah_command[hand] |= SAH_COMMAND_SET_STIFFNESS_FACTOR;
    // impedance control mode
    SDI_F->sah_params[hand].sap_controller = SAH_CONTROLLER_IMPEDANCE;
    SDI_F->sah_command[hand] |= SAH_COMMAND_SET_CONTROLLER;
    /* to limit grasping time and avoid infinite loop */
    SDI_F->sah_closei[hand].nbcall = 0;
  }
  return EXEC;
}

/* sahandGraspBottleMain  -  codel EXEC of GraspBottle
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandGraspBottleMain(int *report)
{
  double *fPara, *f0Para;
  double *fState;
  int finger;
  int hand;
  int nbTighten, nbHandTighten;
  int nbHandActive;
  double fDistal;
  double fProximal;
  SAHAND_HAND_CLOSE *para;
  SAHAND_FINGER_CLOSE* paraFinger;
  double posError, angleError, forceError;
  double Kpos, Kangle, Kforce;
  // 1
  // Pour chaque doigt on cherche le contact
  // 2 
  // Serrage plus energique : position et raideur
  // hypothesis : the additional thumb axis is not used

  if ( SDI_F->sah_closei[hand].nbcall >= 500 ) {
    *report = S_sahand_GRASP_ERROR;
    return ETHER;
  }
  SDI_F->sah_closei[hand].nbcall ++;
  for (hand = 0; hand < 2; hand++) {
    if (!SAHAND_PORT_ACTIVE(hand)) continue;
    switch ( SDI_F->sah_closei[hand].stateHand ) {
    case SAHAND_CLOSING_PROXIMAL:
      for (finger = 0; finger < SAH_MAX_FINGERS; finger++) {
	fPara = &(SDI_F->sah_close[hand].fingers[finger].force[0]);
	f0Para = &(SDI_F->sah_closei[hand].fingersI[finger].force0[0]);
	fState = &(SDI_F->sah_state[hand].ss_finger[finger].sfs_torque[0]);
	if (SDI_F->sah_closei[hand].fingersI[finger].state[SAHAND_PROXIMAL]
	    == SAHAND_FINGER_CLOSING) {
	  fProximal = fabs(fState[SAHAND_PROXIMAL] - f0Para[SAHAND_PROXIMAL]);
	  if ( fProximal >= fPara[SAHAND_PROXIMAL] ) {
	    //printf("fProximal= %g>+%g finger=%d  --%d\n", fProximal,
	    //    fPara[SAHAND_PROXIMAL], finger, SDI_F->sah_closei[hand].nbcall);
	    SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_PROXIMAL] =
	      SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[SAHAND_PROXIMAL];
	    SDI_F->sah_ref[hand].sr_fingers[finger].sfr_vel[SAHAND_PROXIMAL] = 0;
	    SDI_F->sah_params[hand].sap_stiffness[finger].ssf_value[SAHAND_PROXIMAL] =
	      SDI_F->sah_close[hand].fingers[finger].stiffness[SAHAND_PROXIMAL];
	    SDI_F->sah_closei[hand].fingersI[finger].state[SAHAND_PROXIMAL] =
	      SAHAND_FINGER_INACTIVE;	
	    SDI_F->sah_closei[hand].fingersI[finger].state[SAHAND_DISTAL] =
	      SAHAND_FINGER_CLOSING;
	    SDI_F->sah_command[hand] |= SAH_COMMAND_MOVE_HAND;
	    SDI_F->sah_command[hand] |= SAH_COMMAND_SET_STIFFNESS_FACTOR;
	  }
	}
      }
      for (finger = 0, nbTighten = 0; finger < SAH_MAX_FINGERS; finger++) {
	if ( SDI_F->sah_closei[hand].fingersI[finger].state[SAHAND_DISTAL] == 
	     SAHAND_FINGER_CLOSING )
	  nbTighten++;
      }
      if ( nbTighten >= 4) {
	SDI_F->sah_closei[hand].stateHand = SAHAND_CLOSING_DISTAL;
	SDI_F->sah_closei[hand].nbcall = 0;
	printf("SAHAND_CLOSING_DISTAL --%d\n", SDI_F->sah_closei[hand].nbcall);
      }
      break;
    case SAHAND_CLOSING_DISTAL:
      for (finger = 0; finger < SAH_MAX_FINGERS; finger++) {
	paraFinger = &(SDI_F->sah_close[hand].fingers[finger]);
	SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_DISTAL] =
	  paraFinger->positionEnd[SAHAND_DISTAL];
	SDI_F->sah_ref[hand].sr_fingers[finger].sfr_vel[SAHAND_DISTAL] = 
	  paraFinger->velocity[SAHAND_DISTAL];
	SDI_F->sah_params[hand].sap_stiffness[finger].ssf_value[SAHAND_DISTAL] =
	  paraFinger->stiffness[SAHAND_DISTAL];
      }
      SDI_F->sah_command[hand] |= SAH_COMMAND_MOVE_HAND;
      SDI_F->sah_command[hand] |= SAH_COMMAND_SET_STIFFNESS_FACTOR;
      SDI_F->sah_closei[hand].stateHand = SAHAND_TIGHTEN;
      SDI_F->sah_closei[hand].nbcall = 0;
      printf("SAHAND_TIGHTEN --%d\n", SDI_F->sah_closei[hand].nbcall);
      break;
    case SAHAND_TIGHTEN:
      para = &(SDI_F->sah_close[hand]);
      for (finger = 0; finger < SAH_MAX_FINGERS; finger++) {
	if ( ( SDI_F->sah_closei[hand].fingersI[finger].state[SAHAND_DISTAL]
	       == SAHAND_FINGER_CLOSING ) || 
	     ( SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[SAHAND_DISTAL]
	       >= para->fingers[finger].positionEnd[SAHAND_DISTAL] ) ) {
	  fPara = &(SDI_F->sah_close[hand].fingers[finger].force[0]);
	  f0Para = &(SDI_F->sah_closei[hand].fingersI[finger].force0[0]);
	  fState = &(SDI_F->sah_state[hand].ss_finger[finger].sfs_torque[0]);
	  fDistal = fabs(fState[SAHAND_DISTAL] - f0Para[SAHAND_DISTAL]);
	  if ( fDistal >= fPara[SAHAND_DISTAL] ) {
	    //printf("fDistal= %g>+%g finger=%d  --%d\n", fDistal,
	    //  fPara[SAHAND_DISTAL], finger, SDI_F->sah_closei[hand].nbcall);
	    SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_DISTAL] =
	      SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[SAHAND_DISTAL];
	    SDI_F->sah_ref[hand].sr_fingers[finger].sfr_vel[SAHAND_DISTAL] = 0;
	    SDI_F->sah_params[hand].sap_stiffness[finger].ssf_value[SAHAND_DISTAL] =
	      SDI_F->sah_close[hand].fingers[finger].stiffness[SAHAND_DISTAL];
	    SDI_F->sah_closei[hand].fingersI[finger].state[SAHAND_DISTAL] =
	      SAHAND_FINGER_INACTIVE;	
	    SDI_F->sah_command[hand] |= SAH_COMMAND_MOVE_HAND;
	    SDI_F->sah_command[hand] |= SAH_COMMAND_SET_STIFFNESS_FACTOR;
	  }
	}
      }
      for (finger = 0, nbTighten = 0; finger < SAH_MAX_FINGERS; finger++) {
	if (SDI_F->sah_closei[hand].fingersI[finger].state[SAHAND_DISTAL]
	    == SAHAND_FINGER_INACTIVE) {
	  nbTighten++;
	}
      }
      if (nbTighten >= 4) {
	SDI_F->sah_closei[hand].stateHand = SAHAND_GRASPING;
	SDI_F->sah_closei[hand].nbcall = 0;
	printf("SAHAND_GRASPING  --%d\n", SDI_F->sah_closei[hand].nbcall);
      }
      break;
    case SAHAND_GRASPING:
      for (finger = 0, nbTighten = 0; finger < SAH_MAX_FINGERS; finger++) {
	SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_PROXIMAL] =
	  SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[SAHAND_PROXIMAL] + 
	  SDI_F->sah_close[hand].fingers[finger].positionDelta[SAHAND_PROXIMAL];
	if (SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_PROXIMAL]<
	    SAH_MIN_PROXIMAL_ANGLE) 
	  SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_PROXIMAL] =
	    SAH_MIN_PROXIMAL_ANGLE;
	SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_DISTAL] =
	  SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[SAHAND_DISTAL] + 
	  SDI_F->sah_close[hand].fingers[finger].positionDelta[SAHAND_DISTAL];
	if (SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_DISTAL] <
	    SAH_MIN_DISTAL_ANGLE) 
	  SDI_F->sah_ref[hand].sr_fingers[finger].sfr_angle[SAHAND_DISTAL] =
	    SAH_MIN_DISTAL_ANGLE;
	SDI_F->sah_ref[hand].sr_fingers[finger].sfr_vel[SAHAND_PROXIMAL] =
	  SDI_F->sah_close[hand].fingers[finger].velocityClose[SAHAND_PROXIMAL];
	SDI_F->sah_ref[hand].sr_fingers[finger].sfr_vel[SAHAND_DISTAL] =
	  SDI_F->sah_close[hand].fingers[finger].velocityClose[SAHAND_DISTAL];
	SDI_F->sah_params[hand].sap_stiffness[finger].ssf_value[SAHAND_PROXIMAL] =
	  SDI_F->sah_close[hand].fingers[finger].stiffnessClose[SAHAND_PROXIMAL];
	SDI_F->sah_params[hand].sap_stiffness[finger].ssf_value[SAHAND_DISTAL] =
	  SDI_F->sah_close[hand].fingers[finger].stiffnessClose[SAHAND_DISTAL];
      }
      SDI_F->sah_closei[hand].stateHand = SAHAND_GRASPED;
      for (finger = 0; finger < SAH_MAX_FINGERS; finger++) {
	SDI_F->sah_closei[hand].fingersI[finger].positionGrasped[SAHAND_PROXIMAL] = 
	  SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[SAHAND_PROXIMAL];
	SDI_F->sah_closei[hand].fingersI[finger].positionGrasped[SAHAND_DISTAL] = 
	  SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[SAHAND_DISTAL];
	SDI_F->sah_closei[hand].fingersI[finger].positionGrasped[SAHAND_BASE] = 
	  SDI_F->sah_state[hand].ss_finger[finger].sfs_pos[SAHAND_BASE];
      }
      SDI_F->sah_command[hand] |= SAH_COMMAND_MOVE_HAND;
      SDI_F->sah_command[hand] |= SAH_COMMAND_SET_STIFFNESS_FACTOR;
      SDI_F->sah_closei[hand].nbcall = 0;
      printf("SAHAND_GRASPED  --%d\n", SDI_F->sah_closei[hand].nbcall);
      break;
    case SAHAND_GRASPED:
      SDI_F->sah_closei[hand].nbcall = 0; // infinite grasp !
      // control of grasp
      //   - position : thumb position
      //   - orientation : ((ring - first) *fingerLength) / distance(approximation)
      //       We suppose fingerLength and distance are identical
      //   - force = thumb force
      posError = 
	SDI_F->sah_state[hand].ss_finger[SAH_FINGER_THUMB].sfs_pos[SAHAND_PROXIMAL] 
	- SDI_F->sah_closei[hand].fingersI[SAH_FINGER_THUMB].positionGrasped[SAHAND_PROXIMAL];
      angleError =
	( ( SDI_F->sah_state[hand].ss_finger[SAH_FINGER_RING].sfs_pos[SAHAND_PROXIMAL] 
	    - SDI_F->sah_closei[hand].fingersI[SAH_FINGER_RING].positionGrasped[SAHAND_PROXIMAL])
	  -(SDI_F->sah_state[hand].ss_finger[SAH_FINGER_FIRST].sfs_pos[SAHAND_PROXIMAL] 
	    - SDI_F->sah_closei[hand].fingersI[SAH_FINGER_FIRST].positionGrasped[SAHAND_PROXIMAL]));
      fProximal = SDI_F->sah_state[hand].ss_finger[SAH_FINGER_THUMB].sfs_torque[SAHAND_PROXIMAL];
      forceError = fProximal
	- SDI_F->sah_close[hand].fingers[SAH_FINGER_THUMB].force[SAHAND_PROXIMAL];
      Kpos = SDI_F->sah_close[hand].fingers[0].Kpos; // we only use thumb parameters
      Kangle = SDI_F->sah_close[hand].fingers[0].Kangle;
      Kforce = SDI_F->sah_close[hand].fingers[0].Kforce;
      SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_THUMB].sfr_angle[SAHAND_PROXIMAL] =
	SDI_F->sah_state[hand].ss_finger[SAH_FINGER_THUMB].sfs_pos[SAHAND_PROXIMAL]
	+ Kpos * posError + 0.0 * angleError - Kforce * forceError;
      SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_FIRST].sfr_angle[SAHAND_PROXIMAL] =
	SDI_F->sah_state[hand].ss_finger[SAH_FINGER_FIRST].sfs_pos[SAHAND_PROXIMAL]
	- Kpos * posError + Kangle * angleError - Kforce * forceError;
      SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_MIDDLE].sfr_angle[SAHAND_PROXIMAL] =
	SDI_F->sah_state[hand].ss_finger[SAH_FINGER_MIDDLE].sfs_pos[SAHAND_PROXIMAL]
	- Kpos * posError + 0.0 * angleError - Kforce * forceError;
      SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_RING].sfr_angle[SAHAND_PROXIMAL] =
	SDI_F->sah_state[hand].ss_finger[SAH_FINGER_RING].sfs_pos[SAHAND_PROXIMAL]
	- Kpos * posError - Kangle * angleError - Kforce * forceError;
      sahandFingerPositionLimit (&(SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_THUMB].sfr_angle[SAHAND_PROXIMAL]));
      sahandFingerPositionLimit (&(SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_FIRST].sfr_angle[SAHAND_PROXIMAL]));
      sahandFingerPositionLimit (&(SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_MIDDLE].sfr_angle[SAHAND_PROXIMAL]));
      sahandFingerPositionLimit (&(SDI_F->sah_ref[hand].sr_fingers[SAH_FINGER_RING].sfr_angle[SAHAND_PROXIMAL]));
      SDI_F->sah_command[hand] |= SAH_COMMAND_MOVE_HAND;
      printf("SAHAND_TIGHTEN --%d %g %g\n", SDI_F->sah_closei[hand].stateHand,
	     SDI_F->sah_state[hand].ss_finger[SAH_FINGER_THUMB].sfs_pos[1],
	     SDI_F->sah_state[hand].ss_finger[SAH_FINGER_THUMB].sfs_pos[2]);
      break;
    default:
      printf("sahandGraspBottleMain : pb switch %d\n", 
	     SDI_F->sah_closei[hand].stateHand);
      SDI_F->sah_closei[hand].stateHand = SAHAND_FAILED;
      return ETHER;
    }
  }
//  nbHandTighten = nbHandActive = 0;
//  for (hand = 0; hand < 2; hand++) {
//    if (!SAHAND_PORT_ACTIVE(hand)) continue;
//    nbHandActive++;
//    if ( SDI_F->sah_closei[hand].stateHand == SAHAND_GRASPED )
//      nbHandTighten++;
//  }
//  printf("nbHandTighten= %d  nbHandActive= %d\n", nbHandTighten, nbHandActive);
//  if (nbHandTighten < nbHandActive) {
//    return EXEC;
//  } else {
//    return ETHER;
//  }
  return EXEC;
}

/* sahandGraspBottleEnd  -  codel END of GraspBottle
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandGraspBottleEnd(int *report)
{
  // Ne rien faire
  // consigne position  position courante
  /* ... add your code here ... */
  return ETHER;
}

/* sahandGraspBottleInter  -  codel INTER of GraspBottle
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandGraspBottleInter(int *report)
{
  // retour config de base doigt ouvert, mais y a-t-il un risque de 
  // lacher un objet qui aurait t pris ?
  /* ... add your code here ... */
  return ETHER;
}

/*------------------------------------------------------------------------
 * TrackPos
 *
 * Description: 
 *
 * Reports:      OK
 *              S_sahand_NOT_INITIALIZED
 *              S_sahand_POSTER_NOT_FOUND
 *              S_sahand_POSTER_READ_ERROR
 *              S_sahand_READ_ERROR
 */

/* sahandTrackPosStart  -  codel START of TrackPos
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandTrackPosStart(GEN_STRING128 *trackPosterName, int *report)
{
  if (posterFind(trackPosterName->name, &trackPosterId) == ERROR) {
    printf("ERROR: %s poster not found", trackPosterName);
    *report = S_sahand_POSTER_NOT_FOUND;
    return ETHER;
  }
  return EXEC;
}

void sahandPosterSetFinger(SAHAND_FINGER_REF* fingerRef, 
			   GENMANIP_POSITION_SAHAND_FINGER_STR* sahandPoster)
{
  fingerRef->sfr_angle[SAHAND_DISTAL] = sahandPoster->phaDist;
  fingerRef->sfr_angle[SAHAND_PROXIMAL] = sahandPoster->phaProx;
  fingerRef->sfr_angle[SAHAND_BASE] =  sahandPoster->base;
}

void sahandPosterSetHand(SAHAND_REF* sahRef, 
			 GENMANIP_POSITION_SAHAND_STR* sahandPoster)
{
  sahRef->sr_thumb_orientation = sahandPoster->thumbRotation;
  if (sahRef->sr_thumb_orientation < SAH_MIN_THUMB_ORIENTATION)
    sahRef->sr_thumb_orientation = SAH_MIN_THUMB_ORIENTATION;
  if (sahRef->sr_thumb_orientation > SAH_MAX_THUMB_ORIENTATION)
    sahRef->sr_thumb_orientation = SAH_MAX_THUMB_ORIENTATION;

  sahandPosterSetFinger(&(sahRef->sr_fingers[SAH_FINGER_THUMB]),
			&(sahandPoster->thumb));
  sahandPosterSetFinger(&(sahRef->sr_fingers[SAH_FINGER_FIRST]),
			&(sahandPoster->fore));
  sahandPosterSetFinger(&(sahRef->sr_fingers[SAH_FINGER_MIDDLE]),
			&(sahandPoster->middle));
  sahandPosterSetFinger(&(sahRef->sr_fingers[SAH_FINGER_RING]),
			&(sahandPoster->ring));
  sahandFingerPositionLimit(sahRef->sr_fingers[SAH_FINGER_THUMB].sfr_angle);
  sahandFingerPositionLimit(sahRef->sr_fingers[SAH_FINGER_FIRST].sfr_angle);
  sahandFingerPositionLimit(sahRef->sr_fingers[SAH_FINGER_MIDDLE].sfr_angle);
  sahandFingerPositionLimit(sahRef->sr_fingers[SAH_FINGER_RING].sfr_angle);
}

void sahandPosterHandUpdateFinger(SAHAND_FINGER_STATE* state_finger,
				  GENMANIP_POSITION_SAHAND_FINGER_STR *posterFinger)
{
  posterFinger->base    = state_finger->sfs_pos[SAHAND_BASE];
  posterFinger->phaProx = state_finger->sfs_pos[SAHAND_PROXIMAL];
  posterFinger->phaMid  = state_finger->sfs_pos[SAHAND_DISTAL];
  posterFinger->phaDist = state_finger->sfs_pos[SAHAND_DISTAL];
}

void sahandPosterHandUpdate(SAHAND_STATE* state_hand,
			    GENMANIP_POSITION_SAHAND_STR *posterHand)
{
  posterHand->thumbRotation = state_hand->ss_thumb.sts_angle;
  sahandPosterHandUpdateFinger(&(state_hand->ss_finger[SAH_FINGER_THUMB]),
			       &(posterHand->thumb));
  sahandPosterHandUpdateFinger(&(state_hand->ss_finger[SAH_FINGER_FIRST]),
			       &(posterHand->fore));
  sahandPosterHandUpdateFinger(&(state_hand->ss_finger[SAH_FINGER_MIDDLE]),
			       &(posterHand->middle));
  sahandPosterHandUpdateFinger(&(state_hand->ss_finger[SAH_FINGER_RING]),
			       &(posterHand->ring));
}

/* sahandTrackPosMain  -  codel EXEC of TrackPos
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandTrackPosMain(GEN_STRING128 *trackPosterName, int *report)
{
  STATUS status;
  GENMANIP_SAHAND_STR posterPos;
  int hand;

  if (trackPosterId != NULL) {
    status = sahandGENMANIP_SAHAND_STRPosterRead(trackPosterId, &posterPos);
    if (status == ERROR) {
      printf("ERROR: %s poster cannot be read", trackPosterName);
      *report = S_sahand_POSTER_READ_ERROR;
      return ETHER;      
    }
  } else {
    printf("ERROR: %s poster not initialized", trackPosterName);
    *report = S_sahand_POSTER_NOT_FOUND;
    return ETHER ;
  }
  hand = 0;
  if (SAHAND_PORT_ACTIVE(hand)) {
    sahandPosterSetHand(&(SDI_F->sah_ref[hand]), &(posterPos.handRight));
    sahandPosterHandUpdate(&(SDI_F->sah_state[hand]), 
			   &(SDI_F->posterHand.handRight));
    SDI_F->sah_command[hand] |= SAH_COMMAND_MOVE_HAND;	
  }
  hand = 1;
  if (SAHAND_PORT_ACTIVE(hand)) {
    sahandPosterSetHand(&(SDI_F->sah_ref[hand]), &(posterPos.handLeft));
    sahandPosterHandUpdate(&(SDI_F->sah_state[hand]), 
			   &(SDI_F->posterHand.handLeft));
    SDI_F->sah_command[hand] |= SAH_COMMAND_MOVE_HAND;	
  }
  SDI_F->posterHand.cpt++;
  return EXEC;
}


		       

/* sahandTrackPosEnd  -  codel END of TrackPos
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandTrackPosEnd(GEN_STRING128 *trackPosterName, int *report)
{
  return ETHER;
}

/* sahandTrackPosInter  -  codel INTER of TrackPos
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sahandTrackPosInter(GEN_STRING128 *trackPosterName, int *report)
{
  return ETHER;
}


